//------------------------------------------------------------------------------
// Joe Abreu
// BA(HONS)Computer Visualisation & Animation 2008
// Innovations - 3D Painterly Shader
//------------------------------------------------------------------------------

/* float snoise - adjusted perlin noise to range: -1 < x < 1 */
float snoise(float x)
{ 
	return (2*noise((x)) - 1);
}

float scellnoise(float x;float y)
{
	return (2 * cellnoise((x),(y)) - 1);
}

/* float find_t - This function returned the value of t along a line */
float find_t(point A; point B; point pos)
{
	vector AB = B - A; 				// line to compare
	vector AP = pos - A;			// vector from start to pos
	float angle = AB . AP;			// angle between vectors
	
	vector APonAB = AP * cos(angle);	// vector of AP projected onto AB;
	
	float T = length(APonAB) / length(AB);
	
	return T;
}


/* returns the complemntary color */
color complementary_color( color inColor )
{
	
	/* extract color space into individual elements */
	float r = comp( inColor, 0 );
	float g = comp( inColor, 1 );
	float b = comp( inColor, 2 );
	
	/* calculate values  */
	float largest  = r > g ? (r > b ? r : b ) : (g > b ? g : b );
	float smallest = r < g ? (r < b ? r : b ) : (g < b ? g : b );
	float total = largest + smallest;
	
	/* find complementatry hue */
	r = total - r;
	g = total - g;
	b = total - b;
	
	/* return the complementary color */
    return color "rgb" ( r, g, b );
}

surface painterly_shader ( 
						float Ka = 1;
						float Kd = 0.5;
						float Ks = 0.5;
						float roughness = 0.1;
						float impressionist = 0.0;

						float grid_size = 50;
						float stroke_length = 0.6;
						float start_width = 0.2;
						float end_width = 0.2;
						float stipple_frequency = 1; 
						string base_color_map = "";

						color underpaint = 1;
						float color_variation = 0.0;
					)
{
	/* Initialize Useful Variables */
	normal Nf=faceforward(normalize(N),I);
	vector V = normalize(-I);
	color Ct = underpaint;
	float edge_threshold = 0.025; /* edge threshold to interpolate between */
	float i;
	
	for(i=1;i<=grid_size;i+=1.0) {
	
		float cell_s=s*i;
		float cell_t=t*i;

		float ss = mod(cell_s,1);
		float tt = mod(cell_t,1);
		
		point start=point (0.5, (0.5 - stroke_length/2), 0);
		vector up = vector(0.0,stroke_length,0.0);
		point end = start + up;
		
		
		start = rotate(start, cellnoise(cell_s,cell_t)*i * 2 * PI , point( 0.5,0.5,-1 ), point( 0.5,0.5,1 )); //doesnt rotate
		end = rotate(end,  cellnoise(cell_s,cell_t)*i * 2 * PI , point( 0.5,0.5,-1 ), point( 0.5,0.5,1 ));
		point pos=point (ss,tt,0);
	
		/* Set texturing variables */
		
		float line_width; // determine the line width of each stroke
		
		float ratio_t = find_t( start, end, pos ); // gets the correct ratio along the line( start, end )
		line_width = start_width * ratio_t + ( 1 - ratio_t ) * end_width; //get correct width along line
		line_width += (snoise( ratio_t * stipple_frequency )) * (line_width/4); // randomize line width
		
		/* Texturing */
		
		color stroke_color;
		if( base_color_map != "" ) {
			stroke_color = color texture( base_color_map, (floor(s*i)+0.5)/i  , (floor((1-t)*i)+0.5)/i  );
		} else {
			stroke_color = Cs;
		}
		
		/* noise up stroke color */
		stroke_color += scellnoise(cell_s,cell_t)* color_variation;
		
		/* determine the intensity of the shadow colour*/
		color shadow_intensity = (Ka*ambient()+Kd*diffuse(Nf)) + Ks*specular(Nf,V,roughness);
		float intensity = (comp(shadow_intensity,0) + comp(shadow_intensity,1) + comp(shadow_intensity,2)) / 3.0;
		color complementary = complementary_color( stroke_color );
		
		if( impressionist != 0.0 ) {
			stroke_color = intensity*stroke_color + (1 - intensity)*complementary; //use an impressionist shadow color
		} else {
			stroke_color = stroke_color*shadow_intensity; //darken stroke color
		}
		
		/* find out if the sample point is inside or outside the current brush stroke */
		float dist = ptlined(start,end,pos);
		float inside_stroke = 1 - smoothstep(line_width/2-edge_threshold,line_width/2+edge_threshold,dist);

		/* mix the current colour with the colour of the new layer */
		Ct = mix(Ct,stroke_color,inside_stroke);
	
	}
	/* Shading */
	Oi = Os;
	Ci = Oi * Ct;
}
